/*
pix_ps3eye.cpp
flext-based gem object for creating image source from Sony Ps3Eye cam
Copyright (c) 2009 Jean-Yves Gratius (jyg@gumo.fr)

---------------------------------------------------------------------
Using AlexP PS3Eye SDK

http://alexpopovich.wordpress.com/2008/10/02/sony-ps3eye-camera-directshow-capture-source-filter/

------------------------------------------------------------------
Derived from flext tutorials

FLEXT Copyright (c) 2002-2006 Thomas Grill (gr@grrrr.org)

-------------------------------------------------------------------------


*/

// include flext header
#include <flext.h>

// check for appropriate flext version
#if !defined(FLEXT_VERSION) || (FLEXT_VERSION < 500)
#error You need at least flext version 0.5.0
#endif

//  header for GEM compatibility
#include "gem_headers.h"

//  header for Ps3Eye SDK
#include "IPS3Eyelib.h"

//for future multicam support
#define NBMAX_OF_PS3EYE 1

class pix_ps3eye:
	public flext_base
{
	FLEXT_HEADER_S(pix_ps3eye,flext_base,setup)

public:
	// constructor
	pix_ps3eye(int argc,t_atom *argv);
	//destructor
	~pix_ps3eye();


    // table registering how many pix_ps3eye objects are created for each ps3Eye cam
    static int instanciated[];

    // table of PS3EyeLib objects
    static IPS3EyeLib *pCam[];

   // tables registering capture setting for each ps3eye cam
    static int height[];
    static int width[];
    static int fps[];



protected:
    void gem_state(int argc, t_atom *argv);
    void set_fps(int i);
    void set_resolution(int argc,t_atom *argv);

    void clear_buffer(PBYTE myBuffer, long size)
        {
        long index=size;
        while (index > 0)
            {
            myBuffer[--index] = 0;
            }
    }
    void BGRAtoRGBA(PBYTE myBuffer, int pixel_size)
        {
            BYTE mybyte;
            for (int i=0;i<(pixel_size*4);)
                {
                    mybyte=myBuffer[i];
                    myBuffer[i]=myBuffer[i+2];
                    myBuffer[i+2]=mybyte;
                    i+=4;
                }
        }

    void BGRtoRGB(PBYTE myBuffer, long pixel_size)
            {
            BYTE mybyte;
            for (int i=0;i<(pixel_size*3);)
                {
                    mybyte=myBuffer[i];
                    myBuffer[i]=myBuffer[i+2];
                    myBuffer[i+2]=mybyte;
                    i+=3;
                }
        }


    void set_gain(int i)
        {
        i=min(max(0,i),79);
        pCam[camID]->SetGain(i);
        }

    void set_AGC(int i)
        {
        pCam[camID]->AutoAGC(i==1);
        }

    void set_exposure(int i)
        {
        i=min(max(0,i),400);
        pCam[camID]->SetExposure(i);
        }

    void set_AEC(int i)
        {
        pCam[camID]->AutoAEC(i==1);
        }

    void set_red(int i)
        {
        i=min(max(0,i),255);
        pCam[camID]->GetWhiteBalance(&r, &g, &b);
        pCam[camID]->SetWhiteBalance(i, g, b);
       }

    void set_green(int i)
        {
        i=min(max(0,i),255);
        pCam[camID]->GetWhiteBalance(&r, &g, &b);
        pCam[camID]->SetWhiteBalance(r, i, b);
       }

    void set_blue(int i)
        {
         i=min(max(0,i),255);
        pCam[camID]->GetWhiteBalance(&r, &g, &b);
        pCam[camID]->SetWhiteBalance(r, g, i);
       }

    void set_AWB(int i)
        {
       pCam[camID]->AutoAWB(i==1);
        }

    void set_COLOR_BAR(int i)
        {
       pCam[camID]->ColorBar(i==1);
        }


    void set_colorspace(int argc,t_atom *argv)
        {
            if(argc==1)
                if (IsSymbol(argv[0]))
                    {
                    const t_symbol *s = GetSymbol(argv[0]);
        // matching colorspace name
                    if (s==color_RGBA)
                        {
                            my_pixBlock.image.csize=channels=4;
                            my_pixBlock.image.format=colorspace=GL_RGBA;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                   if (s==color_RGB)
                        {
                            my_pixBlock.image.csize=channels=3;
                            my_pixBlock.image.format=colorspace=GL_RGB;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                    if (s==color_RGBA)
                        {
                            my_pixBlock.image.csize=channels=4;
                            my_pixBlock.image.format=colorspace=GL_RGBA;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                    if (s==color_BGRA)
                        {
                            my_pixBlock.image.csize=channels=4;
                            my_pixBlock.image.format=colorspace=GL_BGRA;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                    if (s==color_BGR)
                        {
                            my_pixBlock.image.csize=channels=3;
                            my_pixBlock.image.format=colorspace=GL_BGR;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                    if (s==color_BGR16)
                        {
                            my_pixBlock.image.csize=channels=2;
                            my_pixBlock.image.format=colorspace=GL_RGB16;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                   if ((s==color_Grey)||(s==color_GREY))
                        {
                            my_pixBlock.image.csize=channels=1;
                            my_pixBlock.image.format=colorspace=GL_LUMINANCE;
                            clear_buffer(pBuffer,width[camID]*height[camID]*channels);
                        }
                     }
        }
   void set_upsidedown(int i)
        {
                my_pixBlock.image.upsidedown = (1- (flip=i));
        }


private:



  //  float arg;  // argument variable


        //image transfert parameters
    int channels; // 1 (gray)  / 3 (BGR)   / 4  (BGRA)....
    int format_index;
    unsigned int colorspace;
    int my_width;
    BYTE r,g,b;

 //  entries in table of symbols for colorspace names
    const t_symbol *color_RGBA, *color_RGB, *color_BGRA, *color_BGR, *color_BGR16, *color_Grey, *color_GREY;


    // index specifying which cam is used by this instance of pix_ps3eye
    int camID;

    // image origin(top-left or bottom-left)
    int flip;


    // image buffer and pix structures
    pixBlock my_pixBlock;
    imageStruct my_imageStruct;
    PBYTE pBuffer;
    GemState *gem_state_pointer;

    t_atom my_output;

    FLEXT_CALLBACK_G(gem_state)
    FLEXT_CALLBACK_I(set_fps)
    FLEXT_CALLBACK_V(set_resolution)

    FLEXT_CALLBACK_I(set_gain)
   FLEXT_CALLBACK_I(set_AGC)
    FLEXT_CALLBACK_I(set_exposure)
    FLEXT_CALLBACK_I(set_AEC)
    FLEXT_CALLBACK_I(set_red)
    FLEXT_CALLBACK_I(set_green)
    FLEXT_CALLBACK_I(set_blue)
    FLEXT_CALLBACK_I(set_AWB)
    FLEXT_CALLBACK_I(set_COLOR_BAR)
    FLEXT_CALLBACK_V(set_colorspace)
    FLEXT_CALLBACK_I(set_upsidedown)


    	//setup function : initialise static tables
	static void setup(t_classid c)
	{
        int i;
        for (i=0;i<NBMAX_OF_PS3EYE;i++)
        {
            instanciated[i]=0;
            height[i]=240;
            width[i]=320;
            fps[i]=30;
        }



    }

};


// instantiate static tables
int pix_ps3eye::instanciated[NBMAX_OF_PS3EYE];
int pix_ps3eye::height[NBMAX_OF_PS3EYE];
int pix_ps3eye::width[NBMAX_OF_PS3EYE];
int pix_ps3eye::fps[NBMAX_OF_PS3EYE];


// instantiate pcam[] table : handles to ps3eye
IPS3EyeLib * pix_ps3eye::pCam[NBMAX_OF_PS3EYE];

// instantiate the class
FLEXT_NEW_V("pix_ps3eye",pix_ps3eye)

// class constructor
pix_ps3eye::pix_ps3eye(int argc,t_atom *argv)
{
	// define inlets
	AddInAnything();  // add inlet of type anything (index 0)
    AddOutAnything();

	// register methods
    FLEXT_ADDMETHOD_(0,"gem_state", gem_state);

    // register method for tag "fps" and int argument
    FLEXT_ADDMETHOD_I(0,"fps",set_fps);

    // register method for tag "color" and variable arguments
    FLEXT_ADDMETHOD_(0,"dimen",set_resolution);

    FLEXT_ADDMETHOD_I(0,"gain",set_gain);
  FLEXT_ADDMETHOD_I(0,"AGC",set_AGC);
    FLEXT_ADDMETHOD_I(0,"exposure",set_exposure);
    FLEXT_ADDMETHOD_I(0,"AEC",set_AEC);
    FLEXT_ADDMETHOD_I(0,"red",set_red);
    FLEXT_ADDMETHOD_I(0,"green",set_green);
    FLEXT_ADDMETHOD_I(0,"blue",set_blue);
    FLEXT_ADDMETHOD_I(0,"AWB",set_AWB);
    FLEXT_ADDMETHOD_I(0,"COLOR_BAR",set_COLOR_BAR);
    FLEXT_ADDMETHOD_(0,"colorspace",set_colorspace);
    FLEXT_ADDMETHOD_I(0,"upsidedown",set_upsidedown);



    // init default values
    channels=4;
    flip=0;
    colorspace=GL_RGBA;
    my_width=width[camID];




          // filling table of symbols with colorspace names
         color_RGBA=gensym("RGBA");
         color_RGB=gensym("RGB");
         color_BGRA=gensym("BGRA");
         color_BGR=gensym("BGR");
         color_BGR16=gensym("BGR16");
         color_Grey=gensym("Grey");
         color_GREY=gensym("GREY");



        // choose which camera to open

    camID=0;
    if(argc >= 1)
            if  (IsFloat(argv[0]))
                camID=GetInt(argv[0]);

    if (camID>=NBMAX_OF_PS3EYE)
        {
            post("error : multicam support limited to %d cam",NBMAX_OF_PS3EYE);
             InitProblem(); //Signal a construction problem
        }


    // allocate space for image buffer (choose max resolution, max channels)
      pBuffer=new BYTE[640*480*4];

    //link gem_structures

    my_pixBlock.image=my_imageStruct;

        // redefine my_pixBlock fields
    my_pixBlock.image.csize= channels;
    my_pixBlock.image.xsize= width[camID];
    my_pixBlock.image.ysize= height[camID];
    my_pixBlock.image.data= (unsigned char*)pBuffer;

        // data type - always GL_UNSIGNED_BYTE (except for OS X)
    my_pixBlock.image.type= GL_UNSIGNED_BYTE;

            // is this owned by us (?)
    my_pixBlock.image.notowned = 1;

    my_pixBlock.image.upsidedown = (1- flip);

    my_pixBlock.image.format=colorspace;





    //Check ps3Eye and Start capture
    if(!instanciated[camID])
        {
           // Query for PS3Eye cameras
            if (camID>=IPS3EyeLib::GetNumCameras())
            {
                 // No cameras installed
                post("Error : PS3Eye #%d not detected", camID);
                InitProblem(); //Signal a construction problem
            }
            else
            {
                //TODO : rewrite for future multicam support
            pCam[camID]=IPS3EyeLib::Create();
            instanciated[camID]=1;
            format_index = pCam[camID]->GetFormatIndex(width[camID],height[camID],fps[camID]);
            if (format_index>=0)
                {
                pCam[camID]->SetFormat(format_index);
                pCam[camID]->StartCapture();
                }
            }
        }
        else
        {
            instanciated[camID]++;
         }


}

//destructor
pix_ps3eye::~pix_ps3eye()
{
    if(instanciated[camID])
        {
            instanciated[camID]--;
            if(!instanciated[camID])
                {
                    pCam[camID]->StopCapture();
                    delete pCam[camID];
                }
            }
    	delete [] pBuffer;
}


void pix_ps3eye::gem_state(int argc, t_atom *argv)
{

if (argc==2)   /// TO DO: forward inlet if argc == 1  ('gem_state 1'  or 'gem_state 0')
    if (CanbePointer(argv[1]))
        {
        gem_state_pointer =  (GemState *) GetPointer(argv[1]);

        // connect gem_pointer to my_pixBlock
        gem_state_pointer->image=&my_pixBlock;
        gem_state_pointer->texCoords=0;

// check if resolution has changed
        if (my_width!=width[camID])
            {
                clear_buffer(pBuffer, channels*width[camID]*height[camID]);
                my_width=width[camID];
                my_pixBlock.image.xsize= width[camID];
                my_pixBlock.image.ysize= height[camID];
            }


        // copy image from ps3eye driver
        pCam[camID]->GetFrame(pBuffer, 8*channels, flip, 0);

        // check for colorspace conversion
        if (colorspace==GL_RGBA)
            BGRAtoRGBA(pBuffer,width[camID]*height[camID]);
        else
            if (colorspace==GL_RGB)
            BGRtoRGB(pBuffer,width[camID]*height[camID]);
        my_pixBlock.newimage=1;

     // forward gem_state pointers to outlet
        ToOutAnything(0,gensym("gem_state"),2,argv);
        }
}

void pix_ps3eye::set_fps(int i)
{
    fps[camID]=i;
    format_index = pCam[camID]->GetFormatIndex(width[camID],height[camID],fps[camID]);
    if (format_index>=0)
        {
            pCam[camID]->StopCapture();
            pCam[camID]->SetFormat(format_index);
            pCam[camID]->StartCapture();
        }
}

void pix_ps3eye::set_resolution (int argc,t_atom *argv)
{
    int new_width, new_height;
    bool result=false;
    if (argc==2)
        if (IsFloat(argv[0])&&IsFloat(argv[1]))
           {
               new_width=GetInt(argv[0]);
               new_height=GetInt(argv[1]);
               result=(( new_width==320)&&(new_height==240))||(( new_width==640)&&(new_height==480));
           }
    if (!result)
        post("pix_ps3eye error : incorrect dimen parameters");
    else
        {
        width[camID]=new_width;
        height[camID]=new_height;
        format_index = pCam[camID]->GetFormatIndex(width[camID],height[camID],fps[camID]);
        if (format_index>=0)
            {
                pCam[camID]->StopCapture();
                pCam[camID]->SetFormat(format_index);
                pCam[camID]->StartCapture();
            }
        }

}




